import pytest
from requests_mock import Mocker

from vesta.chars import COLS
from vesta.chars import ROWS
from vesta.client import Client


@pytest.fixture
def client():
    return Client("key", "secret")


def test_base_url():
    base_url = "https://www.example.com"
    client = Client("key", "secret", base_url=base_url)
    assert client.session.base_url == base_url
    assert base_url in repr(client)


def test_headers():
    client = Client("key", "secret", headers={"User-Agent": "Vesta"})
    assert client.session.headers["X-Vestaboard-Api-Key"] == "key"
    assert client.session.headers["X-Vestaboard-Api-Secret"] == "secret"
    assert client.session.headers["User-Agent"] == "Vesta"


def test_get_subscriptions(client: Client, requests_mock: Mocker):
    subscriptions = [True]
    requests_mock.get(
        "https://platform.vestaboard.com/subscriptions",
        json={"subscriptions": subscriptions},
    )
    assert client.get_subscriptions() == subscriptions


def test_get_viewer(client: Client, requests_mock: Mocker):
    viewer = {"_id": 1}
    requests_mock.get("https://platform.vestaboard.com/viewer", json=viewer)
    assert client.get_viewer() == viewer


def test_post_message_text(client: Client, requests_mock: Mocker):
    text = "abc"

    def match_json(request) -> bool:
        return request.json() == {"text": text}

    requests_mock.post(
        "https://platform.vestaboard.com/subscriptions/sub_id/message",
        additional_matcher=match_json,
        json={},
    )
    client.post_message("sub_id", text)


def test_post_message_list(client: Client, requests_mock: Mocker):
    chars = [[0] * COLS] * ROWS

    def match_json(request) -> bool:
        return request.json() == {"characters": chars}

    requests_mock.post(
        "https://platform.vestaboard.com/subscriptions/sub_id/message",
        additional_matcher=match_json,
        json={},
    )
    client.post_message("sub_id", chars)


def test_post_message_list_dimensions(client: Client):
    with pytest.raises(ValueError, match=rf"expected a \({ROWS}, {COLS}\) array"):
        client.post_message("sub_id", [])


def test_post_message_type(client: Client):
    with pytest.raises(TypeError, match=r"unsupported message type"):
        client.post_message("sub_id", True)  # type: ignore



# Copyright (c) 2021-2022 Jon Parise <jon@indelible.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union
from urllib.parse import urljoin

import requests

from .chars import COLS
from .chars import ROWS
from .chars import Rows

__all__ = ["Client"]


class Session(requests.Session):
    def __init__(self, base_url: str):
        super().__init__()
        self.base_url = base_url

    def request(
        self,
        method: Union[str, bytes],
        url: Union[str, bytes],
        *args,
        **kwargs,
    ) -> requests.Response:
        url = urljoin(self.base_url, url if isinstance(url, str) else url.decode())
        return super().request(method, url, *args, **kwargs)


class Client:
    """Provides a Vestaboard API client interface.

    Credentials must be provided as an ``api_key`` and ``api_secret``.

    Optional, an alternate ``base_url`` can be specified, as well as any
    additional HTTP ``headers`` that should be sent with every request
    (such as a custom `User-Agent` header).
    """

    def __init__(
        self,
        api_key: str,
        api_secret: str,
        *,
        base_url: str = "https://platform.vestaboard.com",
        headers: Optional[Mapping[str, str]] = None,
    ):
        self.session = Session(base_url)
        self.session.headers.update(
            {
                "X-Vestaboard-Api-Key": api_key,
                "X-Vestaboard-Api-Secret": api_secret,
            }
        )
        if headers:
            self.session.headers.update(headers)

    def __repr__(self):
        return f"{type(self).__name__}(base_url={self.session.base_url!r})"

    def get_subscriptions(self) -> List[Dict[str, Any]]:
        """Lists all subscriptions to which the viewer has access."""
        r = self.session.get("/subscriptions")
        r.raise_for_status()
        return r.json().get("subscriptions", [])

    def get_viewer(self) -> Dict[str, Any]:
        """Describes the currently authenticated viewer."""
        r = self.session.get("/viewer")
        r.raise_for_status()
        return r.json()

    def post_message(
        self,
        subscription_id: str,
        message: Union[str, Rows],
    ) -> Dict[str, Any]:
        """Post of a new message to a subscription.

        The authenticated viewer must have access to the subscription.

        `message` can be either a string of text or a two-dimensional (6, 22)
        array of character codes representing the exact positions of characters
        on the board.

        If text is specified, lines will be centered horizontally and
        vertically if possible. Character codes will be inferred for
        alphanumeric and punctuation, or can be explicitly specified in-line in
        the message with curly braces containing the character code.

        :raises ValueError: if `message` is a list with unsupported dimensions
        """
        data: Dict[str, Union[str, Rows]]
        if isinstance(message, str):
            data = {"text": message}
        elif isinstance(message, list):
            if len(message) != ROWS or not all(len(row) == COLS for row in message):
                raise ValueError(
                    f"expected a ({ROWS}, {COLS}) array of encoded characters"
                )
            data = {"characters": message}
        else:
            raise TypeError(f"unsupported message type: {type(message)}")

        r = self.session.post(
            f"/subscriptions/{subscription_id}/message",
            json=data,
        )
        r.raise_for_status()
        return r.json()
